

set -ex



ffmpeg --help
ffmpeg -codecs | grep "DEVI.S zlib"
ffmpeg -codecs | grep "DEV.LS h264"
ffmpeg -protocols | grep "https"
test -f $PREFIX/lib/libavcodec.dylib
test -f $PREFIX/lib/libavdevice.dylib
test -f $PREFIX/lib/libswresample.dylib
test -f $PREFIX/lib/libpostproc.dylib
test -f $PREFIX/lib/libavfilter.dylib
test -f $PREFIX/lib/libswresample.dylib
test -f $PREFIX/lib/libavcodec.dylib
test -f $PREFIX/lib/libavformat.dylib
test -f $PREFIX/lib/libswscale.dylib
test -f $PREFIX/lib/libavresample.dylib
exit 0
